#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string>

#include "types.hpp"
#include "types_isd.hpp"
#include "manager.hpp"
#include "tools.hpp"
#include "tools_isd.hpp"
#include "iSDOS.hpp"
#include "iSDOS_tools.hpp"
#include "iterator.hpp"
#include "lang.hpp"

extern PluginStartupInfo startupInfo;
extern const GUID MainGuid;
void addEndSlash(std::wstring &path);

int Manager::writeFile(const wchar_t *name, const UniHdr& hdr)
{
  u32 attr = FILE_FLAG_SEQUENTIAL_SCAN;
  if(hdr.attr & FLAG_READ_PROTECT  ||
     hdr.attr & FLAG_WRITE_PROTECT ||
     hdr.attr & FLAG_DELETE_PROTECT) attr |= FILE_ATTRIBUTE_READONLY;
  if(hdr.attr & FLAG_HIDDEN)         attr |= FILE_ATTRIBUTE_HIDDEN;

  HANDLE file = CreateFileW(name,
                           GENERIC_WRITE,
                           FILE_SHARE_READ | FILE_SHARE_WRITE,
                           NULL,
                           CREATE_ALWAYS,
                           attr,
                           NULL);
  if(file == INVALID_HANDLE_VALUE)
  {
    const wchar_t *msgItems[4];
    msgItems[0] = GetMsg(MError);
    msgItems[1] = GetMsg(MCanNotCreate);
    msgItems[3] = GetMsg(MOk);

    wchar_t shortName[51];
    msgItems[2] = cropLongName(shortName, name, 50);

    messageBox(FMSG_WARNING /*| FMSG_DOWN*/,
               msgItems,
               sizeof(msgItems)/sizeof(msgItems[0]),
               1);
    return 0;
  }

  u32 size = getSize(hdr);
  if(size != 0)
  {
    u8  buf[blockSize];
    u32 noBytesWritten;
    
    Iterator cur(filer, img, hdr), end;
    while(cur != end)
    {
      readBlock(*cur, buf);
      ++cur;
      if(cur != end)
        WriteFile(file, buf, blockSize, &noBytesWritten, NULL);
      else
      {
        int s = (size%blockSize) ? size%blockSize : blockSize;
        WriteFile(file, buf, s, &noBytesWritten, NULL);
      }
    }
  }
  FILETIME modificationTime{};
  DosDateTimeToFileTime(hdr.date, hdr.file.time, &modificationTime);
  LocalFileTimeToFileTime(&modificationTime, &modificationTime);

  u16        tmp = hdr.file.loadAddr;
  SYSTEMTIME laddr;
  laddr.wMilliseconds = 0;
  laddr.wSecond       = 0;
  laddr.wMinute       = tmp%60;
  tmp /= 60;
  laddr.wHour         = tmp%24;
  tmp /= 24;
  laddr.wDay          = tmp%31 + 1;
  tmp /= 31;
  laddr.wMonth        = tmp+1;
  laddr.wYear         = 2000;

  FILETIME   creationTime;
  SystemTimeToFileTime(&laddr, &creationTime);
  SetFileTime(file, &creationTime, NULL, &modificationTime);
  
  CloseHandle(file);
  return 1;
}

static int isFileExist(const wchar_t* name, const UniHdr& hdr, Action& action)
{
  //   
  WIN32_FIND_DATAW data;
  HANDLE h = FindFirstFileW(name, &data);
  if(h != INVALID_HANDLE_VALUE)
  {
    FindClose(h);
    //     
    if(action == SKIP_ALL) return 0;
    if(action == ASK_USER)
    {
      const wchar_t *msgItems[12];
      msgItems[0] = GetMsg(MWarning);
      msgItems[1] = GetMsg(MAlreadyExists);
      msgItems[3] = L"\x001";
      msgItems[6] = L"\x001";
      msgItems[7] = GetMsg(MOverwrite);
      msgItems[8] = GetMsg(MAll);
      msgItems[9] = GetMsg(MSkip);
      msgItems[10] = GetMsg(MSkipAll);
      msgItems[11] = GetMsg(MCancel);

      wchar_t shortName[50];
      msgItems[2] = cropLongName(shortName, name, 47);

      wchar_t dest  [50];
      wchar_t source[50];

      SYSTEMTIME st = makeDate(hdr);
      wsprintfW(source, GetMsg(MSrc),
               getSize(hdr),
               st.wDay, st.wMonth, st.wYear,
               st.wHour, st.wMinute, st.wSecond);

      FILETIME lastWriteTime;
      FileTimeToLocalFileTime(&data.ftLastWriteTime, &lastWriteTime);

      SYSTEMTIME time;
      FileTimeToSystemTime(&lastWriteTime, &time);
                     
      wsprintfW(dest, GetMsg(MDest),
               data.nFileSizeLow,
               time.wDay, time.wMonth, time.wYear,
               time.wHour, time.wMinute, time.wSecond);
      
      msgItems[4] = source;
      msgItems[5] = dest;
      int askCode = messageBox(FMSG_WARNING /*| FMSG_DOWN*/,
                               msgItems,
                               sizeof(msgItems)/sizeof(msgItems[0]), 5);
      switch(askCode)
      {
        case -1:
        case 4:
                  return -1;
        case 3:
                  action = SKIP_ALL;
        case 2:
                  return 0;
        case 1:
                  action = DO_ALL;
      }
    }
  }
  return 1;
}

int Manager::getOneFile(UniHdr&       h,
                        UniHdr*       pDir,
                        const wchar_t*   to_path,
                        const wchar_t*   from_path,
                        int           move,
                        Action&       action,
                        int           opMode)
{
  wchar_t name[8+3+2];
  makeName(h, name);

  wchar_t fullToName[300];
  makeFullName(fullToName, to_path, name);
  
  wchar_t fullFromName[300];
  makeFullName(fullFromName, from_path, name);
  
  int result = 1;
  if(isDir(h))
  {
    addEndSlash(fullToName);
    CreateDirectoryW(fullToName, NULL);

    addEndSlash(fullFromName);

    UniHdr curDir[128];
    readFolder(h, curDir);

    for(int i = 1; i < curDir[0].dir.totalFiles; ++i)
      if(curDir[i].attr & FLAG_EXIST)
      {
        int tmp_result = getOneFile(curDir[i], curDir, fullToName, fullFromName, move, action, opMode);
        if(tmp_result == 1) continue;

        result = tmp_result;
        if(result == -1) break;
      }
    if(move) writeFolder(curDir);
  }
  else
  {
    if((opMode & OPM_SILENT) == 0)
    {
      const wchar_t *msgItems[5];
      msgItems[3] = GetMsg(MTo);
      if(move)
      {
        msgItems[0] = GetMsg(MMove);
        msgItems[1] = GetMsg(MMoving);
      }
      else
      {
        msgItems[0] = GetMsg(MCopy);
        msgItems[1] = GetMsg(MCopying);
      }

      wchar_t shortfullFromName[45];
      wchar_t shortfullToName[45];
      msgItems[2] = cropLongName(shortfullFromName, fullFromName, 40);
      msgItems[4] = cropLongName(shortfullToName, fullToName, 40);
      messageBox(FMSG_LEFTALIGN, msgItems, sizeof(msgItems)/sizeof(msgItems[0]), 0);
    }

    result = isFileExist(fullToName, h, action);
    if(result == 1) result = writeFile(fullToName, h);
  }

  Action actProtected = DO_ALL;
  Action actFolder    = DO_ALL;

  if(result == 1 && move)
    deleteOneFile(h, pDir, from_path, opMode | OPM_SILENT, actProtected, actFolder);
  return result;
}

int Manager::getFiles(PluginPanelItem *panelItem,
                      int   noItems,
                      int   move,
                      const wchar_t* destPath,
                      int   opMode)
{
  if(noItems == 0) return 1;

  std::wstring DestPathCopy(destPath);

  addEndSlash(DestPathCopy.data());

  wchar_t  msg[100] = L"";
  const wchar_t *act[2];
  const wchar_t *actTo[2];
  const wchar_t *button[2];
  act[0] = GetMsg(MCopy);
  act[1] = GetMsg(MMove);
  actTo[0] = GetMsg(MCopyTo);
  actTo[1] = GetMsg(MMoveTo);
  button[0] = GetMsg(MCopyButton);
  button[1] = GetMsg(MMoveButton);
  
  if(noItems == 1)
    wsprintfW(msg, GetMsg(MFileTo), move ? actTo[1] : actTo[0], panelItem[0].FileName);
  else
    wsprintfW(msg, GetMsg(MFilesTo), move ? actTo[1] : actTo[0], noItems);

  FarDialogItem dialogItems[] =
  {
    { DI_DOUBLEBOX,3, 1, 72, 6, { FALSE },           nullptr, nullptr,                                   0, move ? act[1] : act[0] },
    { DI_TEXT,     5, 2,  0, 0, { FALSE },           nullptr, nullptr,                                   0, msg },
    { DI_EDIT,     5, 3, 70, 3,  { TRUE }, L"XiSD_copy_path", nullptr,                         DIF_HISTORY, DestPathCopy.c_str() },
    { DI_TEXT,     3, 4,  0, 0, { FALSE },           nullptr, nullptr,        DIF_BOXCOLOR | DIF_SEPARATOR, L"" },
    { DI_BUTTON,   0, 5,  0, 0, { FALSE },           nullptr, nullptr, DIF_CENTERGROUP | DIF_DEFAULTBUTTON, move ? button[1] : button[0] },
    { DI_BUTTON,   0, 5,  0, 0, { FALSE },           nullptr, nullptr,                     DIF_CENTERGROUP, GetMsg(MCancel) }
  };

  //    ,  
  if((opMode & OPM_SILENT) == 0)
  {
      std::unique_ptr<void, decltype(SafeDlgFree) *> Dlg{ startupInfo.DialogInit(&MainGuid,
                                       &__uuidof(struct __declspec(uuid("{0D78314E-70FE-4085-8D9E-167D55478D98}")) TDlgGuid),
                                       -1, -1, 76, 8,
                                       NULL,
                                       dialogItems,
                                       std::size(dialogItems), 0, 0, startupInfo.DefDlgProc, nullptr), SafeDlgFree };
      if(Dlg.get() == INVALID_HANDLE_VALUE)
      {
          return -1;
      }
      intptr_t askCode = startupInfo.DialogRun(Dlg.get());
  
      if(askCode != 4) return -1;
      auto Sz = startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, nullptr);
      DestPathCopy.resize(Sz);

      FarDialogItemData ItemData = { sizeof(FarDialogItemData), DestPathCopy.size(), DestPathCopy.data() };
      startupInfo.SendDlgMessage(Dlg.get(), DM_GETTEXT, 2, &ItemData);
  }
  
  //   ,   
  if(GetFileAttributesW(DestPathCopy.c_str()) == 0xFFFFFFFF)
  {
      for(wchar_t *c = DestPathCopy.data(); *c; c++)
      {
          if(*c != ' ')
          {
              for(; *c; c++)
              {
                  if(*c == '\\')
                  {
                      *c = 0;
                      CreateDirectoryW(DestPathCopy.c_str(), NULL);
                      *c = '\\';
                  }
              }
              CreateDirectoryW(DestPathCopy.c_str(), NULL);
              break;
          }
      }
  }
  if(!DestPathCopy.empty() && DestPathCopy.back() != ':') addEndSlash(DestPathCopy);

  if(!openHostFile()) return 0;
  HANDLE screen = startupInfo.SaveScreen(0, 0, -1, -1);

  int    exitCode = 1;
  Action action   = ASK_USER;

  for(int iNum = 0; iNum < noItems; ++iNum)
  {
    int fNum = findFile(files, panelItem[iNum].FileName);
    if(!fNum) continue;
    
    int result = getOneFile(files[fNum], files, destPath, curDir, move, action, opMode);
    
    if(result == -1)
    {
      exitCode = -1;
      break;
    }
    if(result ==  0)
    {
      exitCode = 0;
      continue;
    }
    panelItem[iNum].Flags ^= PPIF_SELECTED;
  }

  startupInfo.RestoreScreen(screen);

  if(move)
  {
    writeFolder(files);
    write_device_sys();
  }

  closeHostFile();
  return exitCode;
}
